from enum import IntEnum, IntFlag


class ItemAttachmentType(IntEnum):
    None_ = 0
    Character = 1
    Item = 2
    ItemGuid = 3


# https://stackoverflow.com/a/53791136
ItemAttachmentType.None_._name_ = "None"
ItemAttachmentType._member_map_["None"] = ItemAttachmentType.None_


class ItemAttachmentSlot(IntEnum):
    Preview = 0
    Item_Helm = 1
    Item_Body = 2
    Item_Ring1 = 3
    Item_Ring2 = 4
    Item_Amulet = 5
    Item_Accessory1 = 6
    Item_Accessory2 = 7
    Item_Accessory3 = 8
    Item_WeaponMH = 9
    Item_WeaponOH = 10
    Item_Weapon2H = 11
    Rune_Root = 12
    Rune_Base1 = 13
    Rune_Base2 = 14
    Rune_Equip1 = 15
    Rune_Equip2 = 16
    Rune_Equip3 = 17
    Rune_Equip4 = 18
    Rune_Equip5 = 19
    Rune_Equip6 = 20
    Rune_Equip7 = 21
    Rune_Equip8 = 22
    Rune_Equip9 = 23
    Rune_Equip10 = 24
    Rune_Equip12 = 25
    Rune_Equip_13 = 26
    Rune_Equip14 = 27
    Ability_Player1 = 28
    Ability_Player2 = 29
    Ability_Player3 = 30
    Ability_Follower1 = 31
    Ability_Follower2 = 32
    Ability_Ultimate = 33
    Ability_Dodge = 34
    Ability_Jump = 35
    Ability_Proc1 = 36
    Ability_Proc2 = 37
    CasualWear = 38
    Stats_Chargen = 39
    Stats_Class = 40
    Stats_Skills = 41
    Item_LyriumDagger = 42
    Misc_Flashlight = 43
    Misc_Temp1 = 44
    Misc_Temp2 = 45
    Misc_Temp3 = 46
    Misc_Temp4 = 47
    Weapon_1_Primary = 48
    Weapon_1_Secondary = 49
    Weapon_2_Primary = 50
    Weapon_2_Secondary = 51
    Weapon_Defend = 52
    Weapon_Swap = 53
    Weapon_Deathblow = 54
    ADS = 55
    ADS_Primary = 56
    ADS_Secondary = 57
    TargetLock = 58
    TargetLock_Forward = 59
    TargetLock_Back = 60
    Command_Ping = 61
    Command_Radial = 62
    Command_Follower1 = 63
    Command_Follower2 = 64
    MountedWeapon = 65
    DefaultAppearance = 66
    Quiver = 67
    Hair = 68
    PartSlot_Set = 69
    SpecialAppearance = 70
    Rune_Root2 = 71
    Undergarment = 72
    Count = 73
    Dynamic = -2
    None_ = -1


# https://stackoverflow.com/a/53791136
ItemAttachmentSlot.None_._name_ = "None"
ItemAttachmentSlot._member_map_["None"] = ItemAttachmentSlot.None_

ITEM_ATTACHMENT_SLOT_NAMES = [e.name for e in ItemAttachmentSlot]
ITEM_ATTACHMENT_SLOT_VALUES = [e.value for e in ItemAttachmentSlot]


class CharacterGender(IntEnum):
    Male = 0
    Female = 1
    NonBinary = 2


CHARACTER_GENDER_LABELS = [e.name for e in CharacterGender]
CHARACTER_GENDER_VALUES = [e.value for e in CharacterGender]


class LootRarity(IntEnum):
    Rarity_None = 0
    Rarity_Common = 1
    Rarity_Uncommon = 2
    Rarity_Rare = 3
    Rarity_Epic = 4
    Rarity_Legendary = 5
    Rarity_Ancient = 6
    Rarity_Max = 7


LOOT_RARITY_NAMES = [e.name for e in LootRarity]
LOOT_RARITY_VALUES = [e.value for e in LootRarity]


class CollectibleSetFlag(IntFlag):
    NoFlags = 0
    IsCollected = 1 << 1
    IsDiscovered = 1 << 2
    IsDisabled = 1 << 3
    IsViewed = 1 << 4
    IsSecret = 1 << 5


class BWFollowerStateFlag(IntFlag):
    NoFlags = 0
    Unlocked = 1
    Available = 2
    WasInLastQuest = 4
    IsHardened = 8
    IsHeroic = 16
    IsWaiting = 32
    IsDead = 64


class EcoQuestRegisteredStateFlags(IntFlag):
    Unset = 0
    Eligible = 1
    Completed = 2
    MissionUnlocked = 4
    Expired = 8
    Cheated = 128


class CharacterArchetype(IntEnum):
    Fortune = 624386075
    Follower_Solas = 394763556
    Follower_Davrin = 2602884150
    Follower_Varric = 3417468734
    Rogue = 1902731980
    Warrior = 267923513
    Follower_Lucanis = 2143795149
    Desperado = 2714609019
    Follower_Taash = 4131396826
    Warden_Cine = 294481
    Ranger_03 = 3822852109
    Ranger_02 = 1480587723
    Follower_Bellara = 116806840
    Dalish = 291152393
    Warden_Strategy = 2257715964
    Mage = 240491018
    Warden_4 = 2903517207
    Follower_Neve = 1928218134
    Warden_Challenger = 3998641339
    Crow = 2366407241
    Warden_Endurance = 4003900063
    Shadow_Evoker = 1837455073
    Ranger_01 = 3517341798
    Watcher = 2325381541
    Warden_Art = 3723887171
    Follower_Harding = 1326121707
    NullPlayer = 3509394015
    Player_RGZtest = 2930410500
    Warden_Power = 28757921
    Warden_Technique = 1486725849
    Follower_Spite = 1887180846
    Follower_Emmrich = 3734548853


VISIBLE_CHARACTER_ARCHETYPES = {
    624386075,
    394763556,
    2602884150,
    3417468734,
    1902731980,
    267923513,
    2143795149,
    2714609019,
    4131396826,
    294481,
    3822852109,
    1480587723,
    116806840,
    291152393,
    2257715964,
    240491018,
    2903517207,
    1928218134,
    3998641339,
    2366407241,
    4003900063,
    1837455073,
    3517341798,
    2325381541,
    3723887171,
    1326121707,
    3509394015,
    2930410500,
    28757921,
    1486725849,
    1887180846,
    3734548853,
}

KNOWN_CHARACTER_ARCHETYPE_VALUES = [e.value for e in CharacterArchetype]
KNOWN_CHARACTER_ARCHETYPE_LABELS = [e.name for e in CharacterArchetype]



import json
from uuid import UUID

from importlib_resources import files

from .highlevel import *  # noqa: F401,F403
from .persistence import *  # noqa: F401,F403
from .types import *  # noqa: F401,F403

# from data files:
_FILES = files("bw_save_game.data")
ALL_ITEMS = json.loads(_FILES.joinpath("veilguard", "item_list.json").read_text("utf-8"))
ALL_CURRENCIES = json.loads(_FILES.joinpath("veilguard", "currencies.json").read_text("utf-8"))
ALL_SKILL_GRAPHS = json.loads(_FILES.joinpath("veilguard", "skill_graphs.json").read_text("utf-8"))
ALL_COLLECTIBLES = json.loads(_FILES.joinpath("veilguard", "collectibles.json").read_text("utf-8"))

# post-processing for data files:
for item in ALL_ITEMS:
    item["key"] = f"{item['name' or 'NO NAME']} ({item['id']})"
    item["guid"] = UUID(item["guid"])

SKILL_GRAPHS = {g["id"]: g for g in ALL_SKILL_GRAPHS}

COLLECTIBLES = sorted(ALL_COLLECTIBLES, key=lambda s: s["name"])
COLLECTIBLE_LABELS = [s["name"] for s in COLLECTIBLES]



# bw-save-game - BioWare save game tools
# Copyright (C) 2024 Tim Niederhausen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Website: https://github.com/timniederhausen/bw_save_game
# -*- coding: utf-8 -*-
from imgui_bundle import portable_file_dialogs


def ask_for_file_to_open(message, wildcard, default_path=""):
    dlg = portable_file_dialogs.open_file(title=message, filters=wildcard.split("|"), default_path=default_path)
    while not dlg.ready():
        pass
    res = dlg.result()
    if res:
        return res[0]
    return None


def ask_for_file_to_save(message, wildcard, default_path=""):
    dlg = portable_file_dialogs.save_file(title=message, filters=wildcard.split("|"), default_path=default_path)
    while not dlg.ready():
        pass
    return dlg.result()


def show_error(message):
    dlg = portable_file_dialogs.message("Error", message, portable_file_dialogs.choice.ok)
    while not dlg.ready():
        pass



# bw-save-game - BioWare save game tools
# Copyright (C) 2024 Tim Niederhausen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Website: https://github.com/timniederhausen/bw_save_game
# -*- coding: utf-8 -*-
from bw_save_game.persistence import (
    EcoPersistenceKey,
    PersistenceFamilyId,
    RegisteredPersistenceKey,
    parse_persistence_key_string,
    registered_persistence_key,
)


def test_key_serialization_with_uid():
    key_str = "7:Registered:3237998318|uid=122476741|1522966120|0"
    key = parse_persistence_key_string(key_str)

    assert isinstance(key, RegisteredPersistenceKey)
    assert key.version == 7
    assert key.family == PersistenceFamilyId.Registered
    assert key.persona_id == 3237998318
    assert key.definition_id == 1522966120
    assert key.uid == 122476741

    assert str(key) == key_str


def test_key_serialization_without_uid():
    key_str = "7:Registered:3237998318|1910846056|0"
    key = parse_persistence_key_string(key_str)

    assert isinstance(key, RegisteredPersistenceKey)
    assert key.version == 7
    assert key.family == PersistenceFamilyId.Registered
    assert key.persona_id == 3237998318
    assert key.definition_id == 1910846056
    assert key.uid is None

    assert str(key) == key_str


def test_key_serialization_eco():
    key_str = "7:Eco:0|743239629|652756416|4166880828"
    key = parse_persistence_key_string(key_str)

    assert isinstance(key, EcoPersistenceKey)
    assert key.version == 7
    assert key.family == PersistenceFamilyId.Eco
    assert key.first == 0
    assert key.second == 743239629
    assert key.definition_id == 652756416
    assert key.last == 4166880828


def test_static_keys():
    assert str(registered_persistence_key(1647819227)) == "7:Registered:3237998318|1647819227|0"



# bw-save-game - BioWare save game tools
# Copyright (C) 2024 Tim Niederhausen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Website: https://github.com/timniederhausen/bw_save_game
# -*- coding: utf-8 -*-
import argparse
import json
import logging
import sys

from bw_save_game import __version__
from bw_save_game.container import read_save_from_reader, write_save_to_writer
from bw_save_game.db_object import from_raw_dict, to_raw_dict
from bw_save_game.db_object_codec import dumps, loads

__author__ = "Tim Niederhausen"
__copyright__ = "Tim Niederhausen"
__license__ = "GPL-3.0-only"

_logger = logging.getLogger(__name__)


# ---- Python API ----


def csav_to_json(filename, output):
    with open(filename, "rb") as f:
        m, d = read_save_from_reader(f)

    m = loads(m)
    d = loads(d)

    if output == "-":
        json.dump(dict(meta=m, data=d), sys.stdout, indent=2, default=to_raw_dict)
    else:
        with open(output, "w", encoding="utf-8") as f:
            json.dump(dict(meta=m, data=d), f, indent=2, default=to_raw_dict)


def json_to_csav(filename, output):
    with open(filename, "r", encoding="utf-8") as f:
        doc = json.load(f, object_hook=from_raw_dict)

    m = doc["meta"]
    d = doc["data"]

    # re-encode our object tree into a DbObject byte string
    m = dumps(m)
    d = dumps(d)

    if output == "-":
        # TODO: warn if not a binary stream?
        write_save_to_writer(sys.stdout, m, d)
    else:
        with open(output, "wb") as f:
            write_save_to_writer(f, m, d)


# ---- CLI ----


def register_common_args(parser):
    """Parse command line parameters

    Args:
      parser (argparse.ArgumentParser): parser object to add arguments to.

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser.add_argument(
        "--version",
        action="version",
        version=f"bw_save_game {__version__}",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )


def parse_from_bin_args(args):
    """Parse command line parameters

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--help"]``).

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(description="Dump a JSON representation of a BioWare Frostbite save game")
    register_common_args(parser)
    parser.add_argument("input", help="Path to the input save game .csav")
    parser.add_argument("output", nargs="?", default="-", help="Path to the output .json document")
    return parser.parse_args(args)


def parse_from_json_args(args):
    """Parse command line parameters

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--help"]``).

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(description="Re-encode a JSON representation to a BioWare Frostbite save game")
    register_common_args(parser)
    parser.add_argument("input", help="Path to the input save game .json")
    parser.add_argument("output", nargs="?", default="-", help="Path to the output save game .csav")
    return parser.parse_args(args)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S")


def run_to_json():
    args = parse_from_bin_args(sys.argv[1:])
    setup_logging(args.loglevel)
    csav_to_json(args.input, args.output)


def run_to_bin():
    args = parse_from_json_args(sys.argv[1:])
    setup_logging(args.loglevel)
    json_to_csav(args.input, args.output)


if __name__ == "__main__":
    # e.g.     python -m bw_save_game.convert test.csav
    run_to_json()



# bw-save-game - BioWare save game tools
# Copyright (C) 2024 Tim Niederhausen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Website: https://github.com/timniederhausen/bw_save_game
# -*- coding: utf-8 -*-
import json
from io import BytesIO
from pathlib import Path

from bw_save_game import dumps, loads, read_save_from_reader, write_save_to_writer
from bw_save_game.db_object import from_raw_dict, to_raw_dict

_DATA_DIR = Path(__file__).parent / "data"
_ACTUAL_SAVE_GAMES = [
    _DATA_DIR / "correct_romance_1.csav",
    _DATA_DIR / "correct_romance_2.csav",
    _DATA_DIR / "wrong_romance_1.csav",
    _DATA_DIR / "wrong_romance_2.csav",
]


def test_actual_save_games_binary():
    for actual_save_path in _ACTUAL_SAVE_GAMES:
        with open(actual_save_path, "rb") as f:
            meta, data = read_save_from_reader(f)

        intermediary = BytesIO()
        write_save_to_writer(intermediary, meta, data)
        intermediary.seek(0)
        meta2, data2 = read_save_from_reader(intermediary)

        assert meta == meta2
        assert data == data2


def test_actual_save_games_python():
    for actual_save_path in _ACTUAL_SAVE_GAMES:
        with open(actual_save_path, "rb") as f:
            meta, data = read_save_from_reader(f)

        meta_py = loads(meta)
        data_py = loads(data)

        meta2 = dumps(meta_py)
        data2 = dumps(data_py)

        assert meta == meta2
        assert data == data2

        assert meta_py == loads(meta2)
        assert data_py == loads(data2)


def test_actual_save_games_json():
    for actual_save_path in _ACTUAL_SAVE_GAMES:
        with open(actual_save_path, "rb") as f:
            meta, data = read_save_from_reader(f)

        meta_py = loads(meta)
        data_py = loads(data)

        meta_py_json = json.dumps(meta_py, default=to_raw_dict)
        data_py_json = json.dumps(data_py, default=to_raw_dict)

        meta_py = json.loads(meta_py_json, object_hook=from_raw_dict)
        data_py = json.loads(data_py_json, object_hook=from_raw_dict)

        meta2 = dumps(meta_py)
        data2 = dumps(data_py)

        assert meta == meta2
        assert data == data2

        assert meta_py == loads(meta2)
        assert data_py == loads(data2)



# bw-save-game - BioWare save game tools
# Copyright (C) 2024 Tim Niederhausen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Website: https://github.com/timniederhausen/bw_save_game
# -*- coding: utf-8 -*-
import ctypes
import json
import typing
from uuid import UUID, uuid1

from imgui_bundle import icons_fontawesome, imgui

from bw_save_game.db_object import Double, Long, from_raw_dict, to_native, to_raw_dict
from bw_save_game.db_object_codec import (
    double_struct,
    float_struct,
    int32_struct,
    int64_struct,
    uint64_struct,
)
from bw_save_game.ui.widgets import show_searchable_combo_box

# imgui_bundle has automatically generated bindings that mishandle void*
# by requiring capsules instead of the raw bytes the actual imgui API wants.
PyCapsule_Destructor = ctypes.CFUNCTYPE(None, ctypes.py_object)
PyCapsule_New = ctypes.pythonapi.PyCapsule_New
PyCapsule_New.restype = ctypes.py_object
PyCapsule_New.argtypes = (ctypes.c_void_p, ctypes.c_char_p, PyCapsule_Destructor)

_NANOBIND_VOIDP_CAPSULE_TYPE = b"nb_handle"


def _wrap_bytes_for_imgui(buffer: bytearray):
    raw = (ctypes.c_ubyte * len(buffer)).from_buffer(buffer)
    return PyCapsule_New(raw, _NANOBIND_VOIDP_CAPSULE_TYPE, PyCapsule_Destructor(0))


def show_json_editor(label: str, value, size=None):
    changed, new_value = imgui.input_text_multiline(label, json.dumps(value, indent=2, default=to_raw_dict), size)
    error = None
    if changed:
        try:
            return True, json.loads(new_value, object_hook=from_raw_dict)
        except ValueError as e:
            error = repr(e)
    if error is not None:
        imgui.text_colored((0.9, 0.01, 0.01, 1.0), f"ERROR: {error}")
    return False, None


def show_uuid_editor(label: str, value: UUID):
    changed, new_value = imgui.input_text(label, str(value))
    error = None
    if changed:
        try:
            return True, UUID(hex=new_value)
        except ValueError as e:
            error = repr(e)
    if error is not None:
        imgui.text_colored((0.9, 0.01, 0.01, 1.0), f"ERROR: {error}")
    return False, None


def show_numeric_value_editor(obj, key, value):
    # Unfortunately the DbObject type system doesn't save whether an integral type is signed or unsigned
    # For the UI we thus support the whole unsigned range for positive numbers.
    typ = None
    encoded_value = None
    if isinstance(value, int):
        if value < 0:
            encoded_value = int32_struct.pack(value)
            typ = imgui.DataType_.s32
        else:
            encoded_value = int64_struct.pack(value)
            typ = imgui.DataType_.s64
    if isinstance(value, Long):
        if value.value < 0:
            encoded_value = int64_struct.pack(value.value)
            typ = imgui.DataType_.s64
        else:
            encoded_value = uint64_struct.pack(value.value)
            typ = imgui.DataType_.u64
    if isinstance(value, float):
        encoded_value = float_struct.pack(value)
        typ = imgui.DataType_.float
    if isinstance(value, Double):
        encoded_value = double_struct.pack(value.value)
        typ = imgui.DataType_.double

    if typ is None:
        # not numeric
        return False, False

    new_value = bytearray(encoded_value)
    capsule = _wrap_bytes_for_imgui(new_value)
    changed = imgui.input_scalar(f"##{key}", typ, capsule)
    if not changed:
        return True, False  # nothing to do

    if isinstance(value, int):
        if typ == imgui.DataType_.s64:
            obj[key] = int64_struct.unpack(new_value)[0]
        else:
            obj[key] = int32_struct.unpack(new_value)[0]
    if isinstance(value, Long):
        obj[key] = Long(int64_struct.unpack(new_value)[0])
    if isinstance(value, float):
        obj[key] = float_struct.unpack(new_value)[0]
    if isinstance(value, Double):
        obj[key] = Double(double_struct.unpack(new_value)[0])
    return True, True


def show_raw_value_editor(obj: typing.MutableMapping, key, value=None):
    if value is None:
        value = obj[key]

    supported = False
    if isinstance(value, bool):
        # https://github.com/ocornut/imgui/issues/623
        imgui.set_next_item_width(-1)
        changed, new_value = imgui.checkbox(f"##{key}", value)
        if changed:
            obj[key] = new_value
        supported = True

    if not supported:
        # https://github.com/ocornut/imgui/issues/623
        imgui.set_next_item_width(-1)
        supported, changed = show_numeric_value_editor(obj, key, value)

    if not supported and isinstance(value, str):
        # https://github.com/ocornut/imgui/issues/623
        imgui.set_next_item_width(-1)
        changed, new_value = imgui.input_text(f"##{key}", value)
        if changed:
            obj[key] = new_value
        supported = True

    if not supported and isinstance(value, UUID):
        # https://github.com/ocornut/imgui/issues/623
        imgui.set_next_item_width(-1 - 30)
        changed, new_value = show_uuid_editor(f"##{key}", value)
        imgui.same_line()
        if imgui.button(icons_fontawesome.ICON_FA_SYNC):
            changed = True
            new_value = uuid1()
        if imgui.begin_item_tooltip():
            imgui.text("Re-generate GUID")
            imgui.end_tooltip()
        if changed:
            obj[key] = new_value
        supported = True

    if not supported:
        raise TypeError(f"Unsupported type {type(value)}")
    return changed


def show_raw_key_value_editor(obj, key, label=None):
    value = obj[key]
    if label is None:
        label = str(key)

    if isinstance(value, (dict, list)):
        is_open, is_removed = imgui.collapsing_header(label, True, imgui.TreeNodeFlags_.allow_overlap)
        if not is_open:
            return

        imgui.push_id(key)
        imgui.indent()
        if isinstance(value, dict):
            for sub_key in value:
                show_raw_key_value_editor(value, sub_key)
        if isinstance(value, list):
            for sub_key in range(len(value)):
                show_raw_key_value_editor(value, sub_key)
        imgui.unindent()
        imgui.pop_id()
        return

    imgui.push_id(key)
    imgui.columns(2)
    imgui.text(label)
    imgui.next_column()

    changed = show_raw_value_editor(obj, key, value)

    imgui.columns(1)
    imgui.pop_id()
    return changed


def show_key_value_options_editor(
    label: str, obj, key, option_values: list, option_names: list[str], default_option_index: int = 0
):
    value = obj.get(key)
    if value is None:
        value = option_values[default_option_index]
    native_value = to_native(value)

    imgui.push_id(key)

    imgui.columns(2)
    imgui.text(label)
    imgui.next_column()

    current_item = None
    for i, option in enumerate(option_values):
        if option == native_value:
            current_item = i

    if current_item is None:
        current_item = default_option_index

    # https://github.com/ocornut/imgui/issues/623
    imgui.set_next_item_width(-1)
    changed, current_item = show_searchable_combo_box("##combo", option_names, current_item)
    if changed:
        obj[key] = option_values[current_item]

    imgui.columns(1)
    imgui.pop_id()
    return changed



# bw-save-game - BioWare save game tools
# Copyright (C) 2024 Tim Niederhausen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Website: https://github.com/timniederhausen/bw_save_game
# -*- coding: utf-8 -*-
from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


from .container import read_save_from_reader, write_save_to_writer
from .db_object_codec import dumps, loads

__all__ = ["read_save_from_reader", "write_save_to_writer", "dumps", "loads", "__version__"]



# bw-save-game - BioWare save game tools
# Copyright (C) 2024 Tim Niederhausen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Website: https://github.com/timniederhausen/bw_save_game
# -*- coding: utf-8 -*-
import struct
from decimal import Decimal
from io import BytesIO
from uuid import UUID

from .db_object import (
    DbAttachment,
    DbObjectId,
    DbRecordId,
    DbSHA1,
    DbTimespan,
    DbTimestamp,
    Double,
    Long,
    Matrix4x4,
    VarInt,
    Vector4D,
)


class UnknownSerializerError(TypeError):
    def __init__(self, key, value):
        super(UnknownSerializerError, self).__init__(
            f"Unable to serialize: key '{key}' value: {value} type: {type(value)}"
        )


# Thankfully they left that enum-to-string in the shipping build!
TYPE_Eoo = 0x0
TYPE_Array = 0x1
TYPE_Object = 0x2
TYPE_HomoArray = 0x3
TYPE_Null = 0x4
TYPE_ObjectId = 0x5
TYPE_Bool = 0x6
TYPE_String = 0x7
TYPE_Integer = 0x8
TYPE_Long = 0x9
TYPE_VarInt = 0xA
TYPE_Float = 0xB
TYPE_Double = 0xC
TYPE_Timestamp = 0xD
TYPE_RecordId = 0xE
TYPE_GUID = 0xF
TYPE_SHA1 = 0x10
TYPE_Matrix44 = 0x11
TYPE_Vector4 = 0x12
TYPE_Blob = 0x13
TYPE_Attachment = 0x14
TYPE_Timespan = 0x15
TYPE_InternalMax = 0x1F
TYPE_Anonymous = 0x80

# Cache all the struct formats we might use.
int32_struct = struct.Struct("<i")
uint32_struct = struct.Struct("<I")
byte_struct = struct.Struct("<B")
int64_struct = struct.Struct("<q")
uint64_struct = struct.Struct("<Q")
int_char_struct = struct.Struct("<ib")
float_struct = struct.Struct("<f")
double_struct = struct.Struct("<d")
record_id_struct = struct.Struct("<HHH")
vec4_struct = struct.Struct("<ffff")


def decode_varint_leb128(data, base):
    """Reads the next few bytes in a file as LEB128/7bit encoding and returns an integer"""
    result, shift = 0, 0
    while True:
        byte = byte_struct.unpack(data[base : base + 1])[0]
        base += 1
        result |= (byte & 0x7F) << shift
        if byte >> 7 == 0:
            return base, result
        shift += 7


def encode_varint_leb128(value):
    result = bytearray()
    while True:
        slice = value & 0x7F
        value >>= 7

        if value:
            slice |= 0x80

        result.append(slice)
        if not value:
            break
    return result


# https://stackoverflow.com/a/49659900
def zigzag(x: int) -> int:
    return x << 1 if x >= 0 else (-x - 1) << 1 | 1


def zagzig(x: int) -> int:
    assert x >= 0
    sign = x & 1
    return -(x >> 1) - 1 if sign else x >> 1


def encode_string(value):
    value = value.encode("utf-8")
    length = len(value)
    return encode_varint_leb128(length + 1) + struct.pack("<%dsb" % (length,), value, 0)


def encode_cstring(value):
    if not isinstance(value, bytes):
        value = str(value).encode("utf-8")
    if b"\x00" in value:
        raise ValueError("Element names may not include NUL bytes.")
        # A NUL byte is used to delimit our string, accepting one would cause
        # our string to terminate early.
    return value + b"\x00"


def encode_binary(value):
    length = len(value)
    return encode_varint_leb128(length) + value


def encode_prefix(element_type, name: str):
    if not name:
        return byte_struct.pack(element_type | TYPE_Anonymous)
    return byte_struct.pack(element_type) + encode_cstring(name)


def encode_value(name, value, buf, on_unknown=None):
    if isinstance(value, bool):
        buf.write(encode_prefix(TYPE_Bool, name))
        buf.write(byte_struct.pack(value))
    elif isinstance(value, int):
        # Auto-upgrade ints to Long if we need to.
        if value < -(2**31) or value >= 2**32:
            buf.write(encode_prefix(TYPE_Long, name))
            buf.write(uint64_struct.pack(value & (2**64 - 1)))
        else:
            buf.write(encode_prefix(TYPE_Integer, name))
            buf.write(uint32_struct.pack(value & (2**32 - 1)))
    elif isinstance(value, float):
        buf.write(encode_prefix(TYPE_Float, name))
        buf.write(float_struct.pack(value))
    elif isinstance(value, str):
        buf.write(encode_prefix(TYPE_String, name))
        buf.write(encode_string(value))
    elif isinstance(value, bytes):
        buf.write(encode_prefix(TYPE_Blob, name))
        buf.write(encode_binary(value))
    elif isinstance(value, UUID):
        buf.write(encode_prefix(TYPE_GUID, name))
        buf.write(value.bytes_le)
    elif isinstance(value, DbObjectId):
        buf.write(encode_prefix(TYPE_ObjectId, name))
        assert len(value.value) == 12
        buf.write(value.value)
    elif isinstance(value, DbSHA1):
        buf.write(encode_prefix(TYPE_SHA1, name))
        assert len(value.value) == 20
        buf.write(value.value)
    elif isinstance(value, DbAttachment):
        buf.write(encode_prefix(TYPE_Attachment, name))
        assert len(value.hash) == 20
        buf.write(value.hash)
    elif isinstance(value, DbTimestamp):
        buf.write(encode_prefix(TYPE_Timestamp, name))
        buf.write(uint64_struct.pack(value.value))
    elif isinstance(value, DbRecordId):
        buf.write(encode_prefix(TYPE_RecordId, name))
        buf.write(record_id_struct.pack(value.extentId, value.pageId, value.slotId))
    elif isinstance(value, Vector4D):
        buf.write(encode_prefix(TYPE_Vector4, name))
        buf.write(vec4_struct.pack(value.x, value.y, value.z, value.w))
    elif isinstance(value, Matrix4x4):
        buf.write(encode_prefix(TYPE_Matrix44, name))
        buf.write(vec4_struct.pack(value.row1.x, value.row1.y, value.row1.z, value.row1.w))
        buf.write(vec4_struct.pack(value.row2.x, value.row2.y, value.row2.z, value.row2.w))
        buf.write(vec4_struct.pack(value.row3.x, value.row3.y, value.row3.z, value.row3.w))
        buf.write(vec4_struct.pack(value.row4.x, value.row4.y, value.row4.z, value.row4.w))
    elif isinstance(value, DbTimespan):
        buf.write(encode_prefix(TYPE_Timespan, name))
        buf.write(encode_varint_leb128(zigzag(value.length)))
    elif value is None:
        buf.write(encode_prefix(TYPE_Null, name))
    elif isinstance(value, dict):
        buf.write(encode_prefix(TYPE_Object, name))
        buf.write(encode_document(value, on_unknown=on_unknown))
    elif isinstance(value, list) or isinstance(value, tuple):
        buf.write(encode_prefix(TYPE_Array, name))
        buf.write(encode_array(value, on_unknown=on_unknown))
    elif isinstance(value, Decimal):
        buf.write(encode_prefix(TYPE_Double, name))
        buf.write(double_struct.pack(float(value)))
    elif isinstance(value, Long):
        buf.write(encode_prefix(TYPE_Long, name))
        buf.write(uint64_struct.pack(value.value))
    elif isinstance(value, VarInt):
        buf.write(encode_prefix(TYPE_VarInt, name))
        buf.write(encode_varint_leb128(zigzag(value.value)))
    elif isinstance(value, Double):
        buf.write(encode_prefix(TYPE_Double, name))
        buf.write(double_struct.pack(value.value))
    else:
        if on_unknown is not None:
            encode_value(name, on_unknown(value), buf, on_unknown)
        else:
            raise UnknownSerializerError(name, value)


def encode_document(obj, on_unknown=None, with_envelope=True):
    buf = BytesIO()
    for name in iter(obj):
        value = obj[name]
        encode_value(name, value, buf, on_unknown)
    encoded = buf.getvalue()
    encoded_size = len(encoded)
    if with_envelope:
        return encode_varint_leb128(encoded_size + 1) + encoded + byte_struct.pack(TYPE_Eoo)
    return encoded


def encode_array(array, on_unknown=None):
    buf = BytesIO()
    for i in range(0, len(array)):
        value = array[i]
        encode_value(None, value, buf, on_unknown)
    encoded = buf.getvalue()
    encoded_size = len(encoded)
    return encode_varint_leb128(encoded_size + 1) + struct.pack("<%dsb" % (encoded_size,), encoded, 0)


def decode_value(base: int, data: bytes):
    header = byte_struct.unpack(data[base : base + 1])[0]
    element_type = header & TYPE_InternalMax

    decode_name = (header & TYPE_Anonymous) == 0
    if decode_name:
        base_after_name = data.index(0, base + 1) + 1
        name = data[base + 1 : base_after_name - 1].decode("utf-8")
        base = base_after_name
    else:
        name = None
        base = base + 1

    if element_type == TYPE_Array:
        base, value = decode_document(data, base, as_array=True)
    elif element_type == TYPE_Object:
        base, value = decode_document(data, base)
    elif element_type == TYPE_Null:
        value = None

    elif element_type == TYPE_ObjectId:
        value = DbObjectId(data[base : base + 12])
        base = base + 12

    elif element_type == TYPE_Bool:
        value = bool(byte_struct.unpack(data[base : base + 1])[0])
        base += 1

    elif element_type == TYPE_String:
        base, length = decode_varint_leb128(data, base)
        value = data[base : base + length - 1]
        value = value.decode("utf-8")
        base += length

    elif element_type == TYPE_Integer:
        value = uint32_struct.unpack(data[base : base + 4])[0]
        base += 4

    elif element_type == TYPE_Long:
        value = uint64_struct.unpack(data[base : base + 8])[0]
        base += 8
        value = Long(value)

    elif element_type == TYPE_VarInt:
        base, value = decode_varint_leb128(data, base)
        value = VarInt(zagzig(value))

    elif element_type == TYPE_Float:
        value = float_struct.unpack(data[base : base + 4])[0]
        base += 4

    elif element_type == TYPE_Double:
        value = double_struct.unpack(data[base : base + 8])[0]
        base += 8
        value = Double(value)

    elif element_type == TYPE_Timestamp:
        value = uint64_struct.unpack(data[base : base + 8])[0]
        base += 8
        value = DbTimestamp(value)

    elif element_type == TYPE_RecordId:
        value = record_id_struct.unpack(data[base : base + 4])[0]
        base += 6
        value = DbRecordId(*value)

    elif element_type == TYPE_GUID:
        value = UUID(bytes_le=data[base : base + 16])
        base += 16

    elif element_type == TYPE_SHA1:
        value = DbSHA1(data[base : base + 20])
        base += 20

    elif element_type == TYPE_Vector4:
        value = vec4_struct.unpack(data[base : base + 16])
        base += 16
        value = Vector4D(*value)

    elif element_type == TYPE_Matrix44:
        row1 = vec4_struct.unpack(data[base : base + 16])
        base += 16
        row2 = vec4_struct.unpack(data[base : base + 16])
        base += 16
        row3 = vec4_struct.unpack(data[base : base + 16])
        base += 16
        row4 = vec4_struct.unpack(data[base : base + 16])
        base += 16

        value = Matrix4x4(Vector4D(*row1), Vector4D(*row2), Vector4D(*row3), Vector4D(*row4))

    elif element_type == TYPE_Blob:
        base, length = decode_varint_leb128(data, base)
        value = data[base : base + length]
        base += length

    elif element_type == TYPE_Attachment:
        value = DbAttachment(data[base : base + 20])
        base += 20

    elif element_type == TYPE_Timespan:
        base, value = decode_varint_leb128(data, base)
        value = DbTimespan(zagzig(value))

    else:
        raise ValueError(f"Unhandled DbObject type {element_type} at {base}")

    return base, name, value


def decode_document(data, base, as_array=False, with_envelope=True):
    if with_envelope:
        base, length = decode_varint_leb128(data, base)
        end_point = base + length
        if data[end_point - 1] not in ("\0", 0):
            raise ValueError("missing null-terminator in document")
    else:
        end_point = len(data)

    retval = [] if as_array else {}
    while base < end_point - 1:
        base, name, value = decode_value(base, data)
        if as_array:
            retval.append(value)
        else:
            retval[name] = value

    return end_point, retval


def dumps(obj, on_unknown=None):
    return encode_document({None: obj}, with_envelope=False, on_unknown=on_unknown)


def loads(data):
    return decode_document(data, 0, with_envelope=False)[1][None]



# bw-save-game - BioWare save game tools
# Copyright (C) 2024 Tim Niederhausen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Website: https://github.com/timniederhausen/bw_save_game
# -*- coding: utf-8 -*-
import ctypes
import gzip
import typing
import zlib

MAGIC = b"<!--DAS"
CURRENT_FORMAT_VERSION = 2
CRC32_STARTING_VALUE = 0xA018471F


class SaveHeader(ctypes.Structure):
    _pack_ = 8
    _fields_ = [
        # Names come from a possibly unused JSON version of this header that
        # Dragon Age The Veilguard.exe can write.
        ("formatversion", ctypes.c_uint32),
        ("data_checksum", ctypes.c_uint32),
        ("data_length", ctypes.c_uint64),
        ("data_compressed_length", ctypes.c_uint64),
        ("meta_length", ctypes.c_uint64),
        ("meta_compressed_length", ctypes.c_uint64),
        ("meta_checksum", ctypes.c_uint32),
    ]


class SaveLoadingError(ValueError):
    pass


def read_save_from_reader(reader: typing.BinaryIO, expected_save_type: bytes = b"C", strict=True):
    magic = reader.read(7)
    if magic != MAGIC:
        raise SaveLoadingError(f"Invalid magic bytes: {magic} != {MAGIC}")

    save_type = reader.read(1)
    if expected_save_type is not None and save_type != expected_save_type:
        raise SaveLoadingError(f"Invalid save type: ${save_type} != {expected_save_type}")

    header = SaveHeader.from_buffer_copy(reader.read(ctypes.sizeof(SaveHeader)))
    if header.formatversion != CURRENT_FORMAT_VERSION:
        raise SaveLoadingError(f"Invalid format version: {header.formatversion} != {CURRENT_FORMAT_VERSION}")

    meta = reader.read(header.meta_compressed_length)
    data = reader.read(header.data_compressed_length)

    # If we want to be sure, verify the integrity of our compressed data
    if strict:
        actual_meta_checksum = zlib.crc32(meta, CRC32_STARTING_VALUE)
        actual_data_checksum = zlib.crc32(data, CRC32_STARTING_VALUE)

        if header.meta_checksum != actual_meta_checksum:
            raise SaveLoadingError(f"Invalid meta checksum: {header.meta_checksum} != {actual_meta_checksum}")
        if header.data_checksum != actual_data_checksum:
            raise SaveLoadingError(f"Invalid data checksum: {header.data_checksum} != {actual_data_checksum}")

    # all good, so now we just have to decompress
    meta = gzip.decompress(meta)
    data = gzip.decompress(data)

    assert len(meta) == header.meta_length
    assert len(data) == header.data_length

    return meta, data


def write_save_to_writer(
    writer: typing.BinaryIO,
    meta: typing.ByteString,
    data: typing.ByteString,
    save_type: bytes = b"C",
):
    header = SaveHeader()
    header.formatversion = CURRENT_FORMAT_VERSION
    header.data_length = len(data)
    header.meta_length = len(meta)

    meta = gzip.compress(meta)
    data = gzip.compress(data)

    header.data_checksum = zlib.crc32(data, CRC32_STARTING_VALUE)
    header.data_compressed_length = len(data)
    header.meta_checksum = zlib.crc32(meta, CRC32_STARTING_VALUE)
    header.meta_compressed_length = len(meta)

    writer.write(MAGIC)
    writer.write(save_type)
    writer.write(bytes(header))
    writer.write(meta)
    writer.write(data)



# bw-save-game - BioWare save game tools
# Copyright (C) 2024 Tim Niederhausen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Website: https://github.com/timniederhausen/bw_save_game
# -*- coding: utf-8 -*-
from dataclasses import dataclass

from imgui_bundle import imgui


@dataclass
class ComboBoxState:
    search_pattern = ""
    filtered_items: list[int] = None
    shown_this_frame = True


# retained state
used_combo_boxes = {}  # type: dict[int, ComboBoxState]


def clear_unused_retained_data():
    global used_combo_boxes
    used_combo_boxes = {k: v for k, v in used_combo_boxes.items() if v.shown_this_frame}
    for v in used_combo_boxes.values():
        v.shown_this_frame = False


def show_searchable_combo_box(
    label: str,
    items: list[str],
    current_item: int,
):
    ctx = imgui.get_current_context()

    window = imgui.internal.get_current_window()
    if window.skip_items:
        return False, current_item

    preview_value = items[current_item]

    id = window.get_id(label)
    popup_id = imgui.internal.im_hash_str("##ComboPopup", 0, id)

    global used_combo_boxes
    retained_data = used_combo_boxes.get(id)
    if retained_data is None:
        retained_data = ComboBoxState()
    else:
        retained_data.shown_this_frame = True

    search_pattern = retained_data.search_pattern

    is_already_open = imgui.internal.is_popup_open(popup_id, 0)

    if not imgui.begin_combo(label, preview_value, imgui.ComboFlags_.height_largest):
        used_combo_boxes[id] = retained_data

        # Show tooltips in case our preview value is too long for the inline preview
        if imgui.begin_item_tooltip():
            imgui.text(preview_value)
            imgui.end_tooltip()

        return False, current_item

    if not is_already_open:
        search_pattern = ""

    imgui.push_style_color(imgui.Col_.frame_bg, (240 / 255, 240 / 255, 240 / 255, 1))
    imgui.push_style_color(imgui.Col_.text, (0, 0, 0, 1))
    imgui.push_item_width(-imgui.FLT_MIN)

    if not is_already_open:
        imgui.set_keyboard_focus_here()
    changed, new_value = imgui.input_text("##ComboWithFilter_inputText", search_pattern)
    if changed:
        search_pattern = new_value

    imgui.pop_style_color(2)

    is_filtering = is_already_open and search_pattern
    if is_filtering:
        if search_pattern != retained_data.search_pattern or not retained_data.filtered_items:
            filtered_items = [i for i, item in enumerate(items) if search_pattern in item]
            retained_data.filtered_items = filtered_items
        else:
            filtered_items = retained_data.filtered_items
        num_items = len(filtered_items)
        try:
            filtered_current_item = filtered_items.index(current_item)
        except ValueError:
            filtered_current_item = None
    else:
        num_items = len(items)

    retained_data.search_pattern = search_pattern

    actual_current_item = filtered_current_item if is_filtering else current_item
    value_changed = False

    imgui.push_style_var(imgui.StyleVar_.frame_border_size, 0)

    # from ListBox()
    height_in_items_f = min(7, num_items) + 0.25
    height = imgui.get_text_line_height_with_spacing() * height_in_items_f + ctx.style.frame_padding.y * 2.0
    if imgui.begin_child("##ComboWithFilter_itemList", (0.0, height), imgui.ChildFlags_.frame_style):
        clipper = imgui.ListClipper()
        clipper.begin(num_items, imgui.get_text_line_height_with_spacing())
        if actual_current_item is not None:
            clipper.include_item_by_index(actual_current_item)
        while clipper.step():
            for idx in range(clipper.display_start, clipper.display_end):
                imgui.push_id(idx)
                if is_filtering:
                    label = items[filtered_items[idx]]
                else:
                    label = items[idx]

                is_selected = idx == actual_current_item
                if imgui.selectable(label, is_selected)[0]:
                    if is_filtering:
                        current_item = filtered_items[idx]
                    else:
                        current_item = idx
                    value_changed = True
                    imgui.close_current_popup()

                if is_selected:
                    imgui.set_item_default_focus()
                    if not is_already_open:
                        imgui.set_scroll_here_y()

                imgui.pop_id()

    imgui.end_child()

    if imgui.is_key_pressed(imgui.Key.escape):
        value_changed = False
        imgui.close_current_popup()

    imgui.pop_style_var()
    imgui.pop_item_width()
    imgui.end_combo()

    if value_changed:
        imgui.internal.mark_item_edited(ctx.last_item_data.id_)

    used_combo_boxes[id] = retained_data
    return value_changed, current_item
